/*************************************************************************************
* 3D Data Visualization - Multi-Layered Networks
* 
* (Concept and some code borrowed from 3D Tree Map)
* Tested in Processing 3                                    
* 
* Author:    Su Burtner                       
* 
* Modified:  March 4th, 2019
* 
* Purpose:   Create a multi-layered network that also shows relationships between
*            networks.
* 
* Usage:     Mouse over a node to see the title.
*
**************************************************************************************/

/*
  The data is three CSVs of titles and subjects for the top three biggest collections
*/

import peasy.*; //3D Navigation library
import controlP5.*; //GUI library

//------------------------------------------------------------Constants Section
                             
final int LAYER_DIST = 400;  //Distance between each layer in virtual space 

//------------------------------------------------------------Variables Section

boolean presentationMode = false;   // true -> automation 
                                    // false -> no automation

PeasyCam cam; //Camera(3D Nav) variable

ControlP5 cp5; //User Interface 

// Variables for data
Table data;
int rows, cols;
String[][] data_matrix;

ArrayList nodes;//1, nodes2, nodes3;
ArrayList edges;//1, edges2, edges3;

final float margin_to_left = 50;
final float margin_to_top = 50;

// Display options
int nodeX, nodeY, nodeZ;
final int sphereSize = 8;

color nodeColor;
boolean nodeOver = false;

//---------------------------------------------------------------------------------------------
void setup() {
  size(800, 800, P3D);
  
  //Initial camera and user interface 
  cam = new PeasyCam(this, 2000);
  cp5 = new ControlP5(this);
  
  cp5.addCheckBox("checkBox")
     .setPosition(20, 10)
     .setSize(15, 15) 
     .addItem("Presentation Mode", 1)
     .activateAll(); //This function makes the check box selected by default
     
  cp5.setAutoDraw(false);
  
  // Put data into tables
  data = loadTable("mapCollections.csv", "header");
  
  rows = data.getRowCount();
  cols = data.getColumnCount();
  
  data_matrix = new String[rows][cols];

  nodes = new ArrayList();
  edges = new ArrayList();
 
    
  // Adding data to data_matrix
  for (int i=0; i < data_matrix.length; i++) {
    for (int j=0; j < data_matrix[0].length; j++) {
      data_matrix[i][j] = data.getString(i, j);
    }
  }
  
  // Add nodes randomly for each collection
  for(int i=0; i < data_matrix.length; i++) {
    if(data_matrix[i][2].equals("camap")) {
      nodes.add(new Node((int)(Math.random()*(width - 2*margin_to_left) +
        margin_to_left), (int)(Math.random()*(height - 2*margin_to_top) +
        margin_to_top), i, 10, 10, "camap", data_matrix[i][0]));
    } else if(data_matrix[i][2].equals("canf")) {
      nodes.add(new Node((int)(Math.random()*(width - 2*margin_to_left) +
        margin_to_left), (int)(Math.random()*(height - 2*margin_to_top) +
        margin_to_top), i, 10, 10, "canf", data_matrix[i][0]));
    } else if(data_matrix[i][2].equals("camapr")) {
      nodes.add(new Node((int)(Math.random()*(width - 2*margin_to_left) +
        margin_to_left), (int)(Math.random()*(height - 2*margin_to_top) +
        margin_to_top), i, 10, 10, "camapr", data_matrix[i][0]));
    }
   }

    
  // Add edges for titles with shared subjects within the same collection
  for(int i=0; i < data_matrix.length; i++) {
    for(int j=0; j < data_matrix.length; j++) {
      if(data_matrix[i][1].equals(data_matrix[j][1]) &&
        data_matrix[i][2].equals("camap") &&
        data_matrix[j][2].equals("camap")) {
          edges.add(new Edge((Node)nodes.get(i), (Node)nodes.get(j), "camap"));
      } else if(data_matrix[i][1].equals(data_matrix[j][1]) &&
        data_matrix[i][2].equals("canf") &&
        data_matrix[j][2].equals("canf")) {
          edges.add(new Edge((Node)nodes.get(i), (Node)nodes.get(j), "canf"));
      } else if(data_matrix[i][1].equals(data_matrix[j][1]) &&
        data_matrix[i][2].equals("camapr") &&
        data_matrix[j][2].equals("camapr")) {
          edges.add(new Edge((Node)nodes.get(i), (Node)nodes.get(j), "camapr"));
      }
    }
  }
  
  // Add edges between collection for titles with shared subject
  for(int i=0; i < data_matrix.length; i++) {
    for(int j=0; j < data_matrix.length; j++) {
      if(data_matrix[i][1].equals(data_matrix[j][1]) &&
        data_matrix[i][2].equals("camap") &&
        data_matrix[j][2].equals("canf")) {
          edges.add(new Edge((Node)nodes.get(i), (Node)nodes.get(j), "inter1"));
      } else if(data_matrix[i][1].equals(data_matrix[j][1]) &&
        data_matrix[i][2].equals("canf") &&
        data_matrix[j][2].equals("camapr")) {
          edges.add(new Edge((Node)nodes.get(i), (Node)nodes.get(j), "inter2"));
      }
    }
  }
  
  // Later display options
  nodeColor = color(100);
  ellipseMode(CENTER);
  
}

// Node and Edge classes
class Node {
  
  // Atrributes
  int xpos, ypos, id, sidex, sidey, fillR, fillB, fillG;
  String collection, title;
  
  // Constructor
  Node(int x, int y, int i, int sx, int sy, String collection, String title) {
    xpos = x;
    ypos = y;
    id = i;
    sidex = sx;
    sidey = sy;
    this.collection = collection;
    this.title = title;
  }
  
  // Get collection
  String getColl() {
    return(collection);
  }
  
  // Get title
  String getTitle() {
    return(title);
  }
  
  // Get x
  int getX() {
    return(xpos);
  }
  
  // Get y
  int getY() {
    return(ypos);
  }
  
  // Display method
  void display(int fillR, int fillB, int fillG) {
    smooth();
    noStroke();
    fill(fillR, fillB, fillG);
    ellipse(xpos, ypos, sidex, sidey);
  }
  
  // Make sphere instead of ellipse
  void display3D(int fillR, int fillB, int fillG) {
    smooth();
    noStroke();
    fill(fillR, fillB, fillG);
    lights();
    
    pushMatrix();
    translate(xpos, ypos, 0);
    sphere(sphereSize);
    popMatrix();
  }
  
  // Make a different visualization method for node selected
    void displayHighlight() {
    smooth();
    noStroke();
    fill(255);
    lights();
    
    pushMatrix();
    translate(xpos, ypos, 0);
    sphere(sphereSize + 0.5);
    popMatrix();
  }
  
  void displayHighlight(int R, int G, int B) {
    smooth();
    noStroke();
    fill(R, G, B);
    lights();
    
    pushMatrix();
    translate(xpos, ypos, 0);
    sphere(sphereSize + 0.5);
    popMatrix();
  }
    
}


class Edge {
  
  // Attributes
  int idOr, idDest, strokeR, strokeG, strokeB, strokeT;
  String collection;
  
  // Constructors
  Edge(Node or, Node dest, String collection) {
    idOr = or.id;
    idDest = dest.id;
    this.collection = collection;
  }
  
  // Get collection
  String getColl() {
    return(collection);
  }
  
  // Get Node origin
  Node getOr() {
    Node orig = (Node) nodes.get(this.idOr);
    return(orig);
  }
  
  // Get Node destination
  Node getDest() {
    Node dest = (Node) nodes.get(this.idDest);
    return(dest);
  }
  
  // Get Node origin x
  int getOrX() {
    Node orig = (Node) nodes.get(this.idOr);
    int orX = orig.xpos;
    return(orX);
  }
  
  // Get Node origin y
  int getOrY() {
    Node orig = (Node) nodes.get(this.idOr);
    int orY = orig.ypos;
    return(orY);
  }
  
  // Get Node dest x
  int getDestX() {
    Node dest = (Node) nodes.get(this.idDest);
    int destX = dest.xpos;
    return(destX);
  }
  
  // Get Node dest y
  int getDestY() {
    Node dest = (Node) nodes.get(this.idDest);
    int destY = dest.ypos;
    return(destY);
  }
  
  // Display method
  void display(int strokeR, int strokeG, int strokeB, int strokeT) {
    smooth();
    Node orig = (Node) nodes.get(this.idOr);
    Node dest = (Node) nodes.get(this.idDest);
    strokeWeight(2);
    stroke(strokeR, strokeG, strokeB, strokeT);
    line(orig.xpos, 
         orig.ypos, 
         dest.xpos, 
         dest.ypos);
  }
  
  void displayHighlight(){
    smooth();
    Node orig = (Node) nodes.get(this.idOr);
    Node dest = (Node) nodes.get(this.idDest);
    strokeWeight(3);
    stroke(255, 255, 255);
    line(orig.xpos, 
         orig.ypos, 
         dest.xpos, 
         dest.ypos);
  }
}

//----------------------------------------------------------------------------------------------- //

// Draw
void draw(){
  background(0);  
  rotateX(PI/4);
  rotateZ(PI/4);
  
  //Set automation in presentation mode 
  
  if(presentationMode) {
    rotateY(frameCount*0.0015);
  }
  
  // First network layer
  pushMatrix();
  
  textSize(32);
  fill(255);
  text("camap", width, height);
  
  for(int j=0; j < edges.size(); j++) {
    Edge e1 = (Edge) edges.get(j);
    if(e1.getColl() == "camap") e1.display(87, 70, 123, 90);
  }
  
  for(int i=0; i < nodes.size(); i++) {
    Node n1 = (Node) nodes.get(i);
    
    if(n1.getColl() == "camap") n1.display3D(87, 70, 123);
    
    if(overNode(n1.getX(), n1.getY(), 0)) {
      fill(255);
      textSize(14);
      textMode(SHAPE);
      hint(DISABLE_DEPTH_TEST);
      cam.beginHUD();
      text(n1.getTitle(), mouseX, mouseY);
      cam.endHUD();
      hint(ENABLE_DEPTH_TEST);
    }
  }
  
  popMatrix();
  
  
  // Second network layer
  pushMatrix();
    
  translate(0, 0, LAYER_DIST);
  
  textSize(32);
  fill(255);
  text("canf", width, height);

  for(int j=0; j < edges.size(); j++) {
    Edge e2 = (Edge) edges.get(j);
    if(e2.getColl() == "canf") e2.display(112, 248, 186, 90);
  }
  
  for(int i=0; i < nodes.size(); i++) {
    Node n2 = (Node) nodes.get(i);
    
    if(n2.getColl() == "canf") n2.display3D(112, 248, 186);
    
    if(overNode(n2.getX(), n2.getY(), 0)) {
      fill(255);
      textSize(14);
      textMode(SHAPE);
      hint(DISABLE_DEPTH_TEST);
      cam.beginHUD();
      text(n2.getTitle(), mouseX, mouseY);
      cam.endHUD();
      hint(ENABLE_DEPTH_TEST);
    }
  }
  
  popMatrix();
  
  // Third network layer
  
  pushMatrix();
  
  translate(0, 0, LAYER_DIST*2);
  
  textSize(32);
  fill(255);
  text("camapr", width, height);

  for(int j=0; j < edges.size(); j++) {
    Edge e3 = (Edge) edges.get(j);
    if(e3.getColl() == "camapr") e3.display(202, 254, 74, 90);
  }
  
  for(int i=0; i < nodes.size(); i++) {
    Node n3 = (Node) nodes.get(i);

    if(n3.getColl() == "camapr") n3.display3D(202, 254, 74);
    
    if(overNode(n3.getX(), n3.getY(), 0)) {
      fill(255);
      textSize(14);
      textMode(SHAPE);
      hint(DISABLE_DEPTH_TEST);
      cam.beginHUD();
      text(n3.getTitle(), mouseX, mouseY);
      cam.endHUD();
      hint(ENABLE_DEPTH_TEST);
    }
  }
  
  popMatrix();
  
  // Links between first and second networks
  
  pushMatrix();
  
  beginShape(LINES);
  for(int j=0; j < edges.size(); j++) {
    Edge e4 = (Edge) edges.get(j);
    if(e4.getColl() == "inter1") {
      strokeWeight(2);
      stroke(87, 70, 123);
      vertex(e4.getOrX(), e4.getOrY(), 0);
      stroke(112, 248, 186);
      vertex(e4.getDestX(), e4.getDestY(), LAYER_DIST);
    }
  }
  endShape();
  
  popMatrix();
  
  // Links between second and third networks
  
  pushMatrix();
  
  beginShape(LINES);
  for(int j=0; j < edges.size(); j++) {
    Edge e4 = (Edge) edges.get(j);
    if(e4.getColl() == "inter2") {
      strokeWeight(2);
      stroke(112, 248, 186);
      vertex(e4.getOrX(), e4.getOrY(), LAYER_DIST);
      stroke(202, 254, 74);
      vertex(e4.getDestX(), e4.getDestY(), LAYER_DIST*2);
    }
  }
  endShape();
  
  popMatrix();
  
  
  // Draw text bar at top
  cam.beginHUD();
  
  fill(0);
  stroke(0);
  rect(0, 0, width, 35);
  fill(255);
  textAlign(RIGHT);
  textSize(14);
  text("A multi-layered network of map collections at Seattle Public Library", width - 150, 20);
  
  cp5.draw();
  cam.endHUD();
 
}

// Method for presentation mode
void checkBox(float[] a) {
  if(a[0] == 1) presentationMode = false;
  else presentationMode = true;
}

boolean overNode(int nodeX, int nodeY, int nodeZ) {
  PVector mouseXY = new PVector(mouseX, mouseY);
  PVector screenXY = new PVector(screenX(nodeX, nodeY, nodeZ), screenY(nodeX, nodeY, nodeZ));
  float mouseDist = mouseXY.dist(screenXY);
  
  if(mouseDist < sphereSize/2 ) {
    return true;
  } else {
    return false;
  }
}
